/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/


/*
 * modulecall.h
 *
 *        Monitor <-->  Module (kernel driver) interface
 */

#ifndef _MODULECALL_H
#define _MODULECALL_H

#define INCLUDE_ALLOW_VMCORE
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#include "includeCheck.h"

#include "x86.h"
#include "vm_time.h"
#include "vcpuid.h"
#include "vcpuset.h"

/*
 *----------------------------------------------------------------------
 *
 * ModuleCallType --
 *
 *      Enumaration of support calls done by the module
 *
 *----------------------------------------------------------------------
 */

typedef enum ModuleCallTypeTOT {
   MODULECALL_TOT_NONE = 100,
   
   MODULECALL_TOT_INTR,                  
   MODULECALL_TOT_SEMAWAIT,
   MODULECALL_TOT_SEMASIGNAL,
   MODULECALL_TOT_SEMAFORCEWAKEUP,
   MODULECALL_TOT_IPI,                   // hit thread with IPI
   MODULECALL_TOT_SWITCH_TO_PEER,

   /*
    * Return codes for user calls
    */

   MODULECALL_TOT_USERRETURN,
   MODULECALL_TOT_USERTIMEOUT,

   MODULECALL_TOT_GET_RECYCLED_PAGE,
   MODULECALL_TOT_RELEASE_ANON_PAGE,

   MODULECALL_TOT_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeTOT;

typedef enum ModuleCallTypeV45 {
   MODULECALL_V45_NONE = 100,
   
   MODULECALL_V45_INTR,                  
   MODULECALL_V45_SEMAWAIT,
   MODULECALL_V45_SEMASIGNAL,
   MODULECALL_V45_SEMAFORCEWAKEUP,
   MODULECALL_V45_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_V45_USERRETURN,
   MODULECALL_V45_USERTIMEOUT,

   MODULECALL_V45_GET_RECYCLED_PAGE,
   MODULECALL_V45_RELEASE_ANON_PAGE,
   MODULECALL_V45_SET_FASTCLOCKRATE,

   MODULECALL_V45_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV45;

typedef enum ModuleCallTypeV4 {
   MODULECALL_V4_NONE = 100,
   
   MODULECALL_V4_INTR,                  
   MODULECALL_V4_LOCKPAGE,                  
   MODULECALL_V4_UNLOCKPAGE,                  
   MODULECALL_V4_ISMPNLOCKED, 
   MODULECALL_V4_SEMAWAIT,
   MODULECALL_V4_SEMASIGNAL,
   MODULECALL_V4_SEMAFORCEWAKEUP,
   MODULECALL_V4_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_V4_USERRETURN,
   MODULECALL_V4_USERTIMEOUT,

   MODULECALL_V4_GET_RECYCLED_PAGE,
   MODULECALL_V4_RELEASE_ANON_PAGE,

   MODULECALL_V4_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV4;

typedef enum ModuleCallTypeV3 {
   MODULECALL_V3_NONE = 100,            //
   
   /*
    * All I/O interrupts go under MODULECALL_INTR, special-cased
    * in task switching code
    */
   MODULECALL_V3_INTR,                  
   MODULECALL_V3_LOCKPAGE,                  
   MODULECALL_V3_UNLOCKPAGE,                  

   MODULECALL_V3_ISMPNLOCKED, 
   /*
    * calls to user mode
    */
   MODULECALL_V3_USERRETURN,            // return code for USER RPC
   MODULECALL_V3_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV3;

#define MODULECALL_USER_START 300
#define MODULECALL_USER_END   399

#define MODULECALL_CROSS_PAGE_LEN    1
#define MODULECALL_CROSS_PAGE_START_V2		2
#define MODULECALL_CROSS_PAGE_START_GSX1	2
#define MODULECALL_CROSS_PAGE_START_V3		4
#define MODULECALL_CROSS_PAGE_START_V32		4
#define MODULECALL_CROSS_PAGE_START_V321	4
#define MODULECALL_CROSS_PAGE_START_GSX2	6
#define MODULECALL_CROSS_PAGE_START_GSX25	6
#define MODULECALL_CROSS_PAGE_START_V4		6
#define MODULECALL_CROSS_PAGE_START_V45		6
#define MODULECALL_CROSS_PAGE_START_TOT		6

// we don't want it in the final version
//#define USE_BLUE_SCREEN

#define MODULECALL_USERCALL_NONE     300

/*
 * Define VMX86_UCCOST in the makefiles (Local.mk,
 * typically) if you want a special build whose only purpose
 * is to measure the overhead of a user call and its
 * breakdown.
 *
 * WINDOWS NOTE: I don't know how to pass VMX86_UCCOST to
 * the driver build on Windows.  It must be defined by hand.
 *
 * ESX Note: we don't have a crosspage in which to store these
 * timestamps.  Such a feature would perhaps be nice (if we
 * ever tire of the argument that esx does so few usercalls
 * that speed doesn't matter).
 */

#ifdef VMX86_UCCOST

#ifdef VMX86_SERVER

typedef struct UCCostResults {
   uint32 vmksti;
   uint32 vmkcli;
   uint32 ucnop;
} UCCostResults;

#define UCTIMESTAMP(cp, stamp)

#else

typedef struct UCCostResults {
   uint32 htom;
   uint32 mtoh;
   uint32 ucnop;
} UCCostResults;

typedef enum UCCostStamp {
#define UC(x) UCCOST_ ## x,
#include "uccostTable.h"
   UCCOST_MAX
} UCCostStamp;

#define UCTIMESTAMP(cp, stamp) \
             do { (cp)->ucTimeStamps[UCCOST_ ## stamp] = RDTSC(); } while (0)
#endif // VMX86_SERVER
#else

#define UCTIMESTAMP(cp, stamp)

#endif // VMX86_UCCOST

#define INTERRUPT_REDIRECTION_BITMAP_SIZE 32
#define IO_PERMISSION_BITMAP_SIZE 0

#define SWITCHERROR_IN_PANIC    -1
#define SWITCHERROR_NONE        0

#ifdef _WIN32
#define __attribute__(_x)     // NOP
#include <PSHPACK1.H> // 1-byte packing alignment
#endif

typedef struct Context64 {
   uint64 cr3;
   uint64 rax;
   uint64 rcx;
   uint64 rdx;
   uint64 rbx;
   uint64 rsi;
   uint64 rdi;
   uint64 rbp;
   uint64 rsp;
   uint64 r8;
   uint64 r9;
   uint64 r10;
   uint64 r11;
   uint64 r12;
   uint64 r13;
   uint64 r14;
   uint64 r15;
   uint16 cs;
   uint16 ds;
   uint16 ss;
   uint16 es;
   uint16 fs;
   uint16 gs;
   uint64 rip;
   uint64 eflags;
} __attribute__ ((packed)) Context64;

typedef struct ContextInfo {
   DTRWords32 gdtr;
   DTRWords32 ldtr;
   DTRWords32 idtr;
   uint16   tr;
   uint16   _pad0;
   Task     task;   
} __attribute__ ((packed)) ContextInfo;

typedef struct ContextInfo64 {
   DTRWords64 gdtr64;
   DTRWords64 ldtr64;
   DTRWords64 idtr64;
   DTR64      switchGdtr64;
   uint16     tr;
   FarPtr32   jump32Code;       /* &worldswitch_64h_32v_mode_32compat */
   uint16     _pad;
   FarPtr32   jump64Code;       /* &worldswitch_64h_32v_mode_64 */
   uint16     switchDS;
   uint64     farPtr;           /* &worldswitch_64h_32v_mon_switch */
   Context64  context;
   Task64     task;
} __attribute__ ((packed)) ContextInfo64;

#define WS_ALIGN(_s) ((_s + 15) & ~(15))
#define WS_MAX(_a, _b) ((_a) > (_b)) ? (_a) : (_b)

/* The sizes for the following loadable worldswitch code sections are
 * found by running '$VMTREE/support/scripts/worldswitch_sections'
 * after a build of the monitor has completed successfully.
 */
#define HWS_32H_32V WS_ALIGN(0x0000007b)
#define HWS_64H_64V WS_ALIGN(0x0000011e)
#define HWS_64H_32V WS_ALIGN(0x000000d8)
#define VWS_64H_32V WS_ALIGN(0x000000cd)

#define WS_CODE_STD WS_MAX(HWS_32H_32V, HWS_64H_32V)
#define WS_CODE_VMM VWS_64H_32V
#define WS_CODE_64 HWS_64H_64V

typedef struct ContextSwitch {
   uint8  hostXToVmm32[WS_CODE_STD];           /* hws_32h_32v
                                                * -or- hws_64h_32v */
   uint8  vmm32ToHost64[WS_CODE_VMM];          /* vws_64h_32v */
   uint8  vmm64ToHost64[WS_CODE_64];           /* hws_64h_64h */
} ContextSwitch;

/*
 *----------------------------------------------------------------------
 *
 * VMCrossPage --
 *
 *      data structure shared between the monitor and the module
 *      that is used for crossing between the two.
 *      Accessible as vm->cross (kernel module) and CROSS_PAGE
 *      (monitor)
 *
 *      Exactly one page long
 *
 *----------------------------------------------------------------------
 */

typedef struct VMCrossPageTOT {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  
   uint32 _pad[3]; /* Align context switching code on 16 byte boundary. */

   ContextSwitch contextSwitch;

   /* Pad to offset 1024 to appease the P4 trace cache.  Smaller than
    * 1024 causes the trace cache to be flushed each time data is
    * written to the context switch.
    */
   uint8 _trace_pad[1024 - (16 /* alignment */ + sizeof(ContextSwitch))];

   /*
    * The interrupt redirection bitmap, must immediately follow 
    * monContext.task (ASSERT fail otherwise).
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
   uint32 switchL4Off;
   MPN32 crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 
   VM_L4E lmSwitchPE;
   
   /*
    * In x86,
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    *
    * In x86 VM on x86-64 host,
    * In order to make the transition to legacy mode from long mode,
    * we must first turn paging off on the processor. For this we need to
    * be executing code on an identity mapped page (the crosspage). Here
    * is where we keep the physical address of the crosspage. We will use
    * this address to create an identity mapping (a translation for which
    * PA == LA)
    *
    */
   LA32 crosspageLinearAddr;
   
   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeTOT moduleCallType;
   uint32 args[4];
   uint32 retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool runVmm64;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#if defined(VMX86_UCCOST) && !defined(VMX86_SERVER)
   uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   int32 switchError;

   SysenterState vmmSysenter;

   /*
    * Adjustment for machines where the hardware TSC does not run
    * constantly (laptops).  See VMK_SharedData for the ESX analog.
    * Within the monitor, PTSC_Get() = RDTSC() + tscAdjustment.  To
    * handle the case where the TSC slows or stops when the host OS
    * halts the processor using APM or ACPI, tscAdjustment is updated
    * whenever we go to userlevel.  To handle out of sync MP TSCs,
    * tscAdjustment will need to be per-PCPU and thus will need to be
    * updated whenever this VCPU migrates to a different PCPU.
    */
   VmRelativeTS tscAdjustment;

   /*
    * PTSC value and target VCPUSet for the next MonitorPoll callback,
    * as last known by the VCPU that this crosspage belongs to.  To
    * find up-to-date information, vmmon must loop through the
    * crosspages of all the VCPUs in this VM and take the one with the
    * latest time (ugh).  When the time arrives, if a target VCPU
    * thread is in the monitor, it wants to receive a hardware
    * interrupt (e.g., an IPI) as soon as possible; if it has called
    * up to userlevel to halt, it wants to wake up as soon as
    * possible.
    */
   VmAbsoluteTS monitorPollFirstTS;
   VCPUSet monitorPollFirstVCS;
   
   /*
    * Fields used by x86_64.
    *
    * Since we need to share the same VMX and VMM binary for
    * x86 and x86_64, we don't use ifdef to define x86_64
    * specific fields. If we run out of crosspage space,
    * we need to define these structures using unions.
    */
   DescriptorUnion switchGdt[4];
   ContextInfo64 hostContext64;         // replace hostContext
   uint64 hostContext64VA;              // replace hostContextVA
   uint64 hostVA64;                     
   uint64 hostSwitchCR3;

   uint64 switchHostCR4;
   uint64 hostDR[8];
   uint32 hostDRSaved;
   Bool debugRegistersReady;       // DR invariants are active, may touch DRs
   Bool restoreGeneralDetect;      // vmmon must restore DR7_GD

   char _pad4[2];
   ContextInfo64 monContext64;
#ifdef __cplusplus
   static unsigned int crosspageStart(void) { return 6; }
#endif
} __attribute__ ((packed)) VMCrossPageTOT;


#ifdef WIN32
#include <POPPACK.H>
#endif


/*
 * We hardcode these constants here instead of using the sizeof() and
 * offsetof() to catch compiler inconsistencies. This file is compiled
 * once onsite to build the monitor, and then once at the customer site
 * to build the drivers.
 *
 * An alternate scheme, and preferred scheme, would be have the makefile
 * generate them, thereby relieving the developer of the error prone
 * chore of updating them each time VMCrossPage is modified. 
 *
 * Pratap 12/01.
 */
#define MODULECALL_CROSSPAGE_SIZE_TOT (0x914 + UCCOST_CROSSPAGE_SIZE)

#define MODULECALL_CROSSPAGE_LAOFFSET_TOT 0x550

#define CROSSPAGE_VERSION_TOT 	0x1789

#if !defined(VMX86_SERVER) && defined(VMM)
#define CROSS_PAGE  ((VMCrossPage * const) VPN_2_VA(CROSS_PAGE_START))
#define VMM_SWITCH_SHARED_DATA CROSS_PAGE
#endif

#define NULLPAGE_LINEAR_START  (MONITOR_LINEAR_START + \
                                PAGE_SIZE * CPL0_GUARD_PAGE_START)

#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)

/* ---------------------------- VMware 4.5 --------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V45 1008

typedef struct {
   DTRWords32  gdtr;
   DTRWords32  ldtr;
   DTRWords32  idtr;
   uint16 tr;
   uint8 jump[6];
   Task task;   
} __attribute__ ((packed)) ContextInfoV45;

typedef struct {
   uint64 cr3;

   // general registers
   uint64 rax;
   uint64 rcx;
   uint64 rdx;
   uint64 rbx;
   uint64 rsi;
   uint64 rdi;
   uint64 rbp;
   uint64 rsp;

   // segment registers

   uint16 cs;
   uint16 ds;
   uint16 ss;
   uint16 es;
   uint16 fs;
   uint16 gs;

   // rip
   uint64 rip;

} __attribute__ ((packed)) Context64V45;

typedef struct {
   DTRWords64  gdtr64;
   DTRWords64  ldtr64;
   DTRWords64  idtr64;
   uint16 tr;
   DTR64 switchGdtr64;
   FarPtr32 jump32Code;
   uint16   pad1;
   FarPtr32 jump64Code;
   uint16   pad2;
   uint16   switchDS;
   uint64   farPtr;
   Context64V45 context;
} __attribute__ ((packed)) ContextInfo64V45;

typedef struct VMCrossPageV45 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   struct {
   	uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V45];
	uint8 vmm64ToHost64[0];
   } contextSwitch;

   /*
    * The interrupt redirection bitmap, must immediately follow 
    * monContext.task (ASSERT fail otherwise).
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN32 crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * In x86,
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    *
    * In x86 VM on x86-64 host,
    * In oder to make the transition to legacy mode from long mode,
    * we must first turn paging off on the processor. For this we need to
    * be executing code on an identity mapped page (the crosspage). Here
    * is where we keep the phisical address of the crosspage. We will use
    * this address to create an identity mapping (a translation for which
    * PA == LA)
    *
    */
   LA32 crosspageLinearAddr;
   
   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeV45 moduleCallType;
   uint32 args[4];
   uint32 retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#if defined(VMX86_UCCOST) && !defined(VMX86_SERVER)
   uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   int32 switchError;

   /*
    * monUsesSysenter is TURE iff the virtual MSR_SYSENTER_CS is
    * nonzero and the monitor uses the sysenter/sysexit MSRs
    * for fast system calls.
    */
   Bool monUsesSysenter;

   Bool _pad2[3];

   /*
    * Adjustment for machines where the hardware TSC does not run
    * constantly (laptops).  See VMK_SharedData for the ESX analog.
    * Within the monitor, PTSC_Get() = RDTSC() + tscAdjustment.  To
    * handle the case where the TSC slows or stops when the host OS
    * halts the processor using APM or ACPI, tscAdjustment is updated
    * whenever we go to userlevel.  To handle out of sync MP TSCs,
    * tscAdjustment will need to be per-PCPU and thus will need to be
    * updated whenever this VCPU migrates to a different PCPU.
    */
   VmRelativeTS tscAdjustment;

   /*
    * PTSC value and target VCPUSet for the next MonitorPoll callback,
    * as last known by the VCPU that this crosspage belongs to.  To
    * find up-to-date information, vmmon must loop through the
    * crosspages of all the VCPUs in this VM and take the one with the
    * latest time (ugh).  When the time arrives, if a target VCPU
    * thread is in the monitor, it wants to receive a hardware
    * interrupt (e.g., an IPI) as soon as possible; if it has called
    * up to userlevel to halt, it wants to wake up as soon as
    * possible.
    */
   VmAbsoluteTS monitorPollFirstTS;
   VCPUSet monitorPollFirstVCS;
   
   /*
    * Fields used by x86_64.
    *
    * Since we need to share the same VMX and VMM binary for
    * x86 and x86_64, we don't use ifdef to define x86_64
    * specific fields. If we run out of crosspage space,
    * we need to define these structures using unions.
    */
   DescriptorUnion switchGdt[4];
   ContextInfo64V45 hostContext64;         // replace hostContext
   Bool _pad3[2];
   uint64 hostContext64VA;              // replace hostContextVA
   uint64 hostVA64;                     
   uint64 hostSwitchCR3;

   uint64 switchHostCR4;
   uint64 hostDR[8];
   uint32 hostDRSaved;

#ifdef __cplusplus
   static const Bool runVmm64 = 0;
   static char monContext64;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V45_INTR; }
   static unsigned int crosspageStart(void) { return 6; }
   static SysenterState vmmSysenter;
#endif
} __attribute__ ((packed)) VMCrossPageV45;

#define MODULECALL_CROSSPAGE_SIZE_V45 (0x6f4 + UCCOST_CROSSPAGE_SIZE)
#define MODULECALL_CROSSPAGE_LAOFFSET_V45 0x54c

/* ---------------------------- VMware 4 --------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V4 1008

typedef struct VMCrossPageV4 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V4];
   } contextSwitch;

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   VA32 hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN32 crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA32 crosspageLinearAddr;

   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeV4 moduleCallType;
   uint32 args[4];
   uint32 retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#ifdef VMX86_UCCOST
    uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   uint32 switchPanic;
   Bool monUsesSysenter;

#ifdef __cplusplus
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V4_INTR; }
   Bool isDRModified(int reg, uint32 savedDR) { return (savedDR != shadowDR[reg]); }
   void setDR(int reg, uint32 dr) { shadowDR[reg] = dr; }
   static unsigned int crosspageStart(void) { return 6; }
#endif
} __attribute__ ((packed)) VMCrossPageV4;

#define MODULECALL_CROSSPAGE_SIZE_V4 (0x629 + UCCOST_CROSSPAGE_SIZE)
#define MODULECALL_CROSSPAGE_LAOFFSET_V4 0x5d0

/* ---------------------------- GSX 2.5 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_GSX25  1008

typedef struct VMCrossPageGSX25 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_GSX25];
   } contextSwitch;

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   VA32 hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN32 crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   LA32 monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA32 crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

#ifdef __cplusplus
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 6; }
#endif
} __attribute__ ((packed)) VMCrossPageGSX25;

#define MODULECALL_CROSSPAGE_SIZE_GSX25     0x5fd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX25 0x5d4

/* ---------------------------- GSX 2.0 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_GSX2  1008

typedef struct VMCrossPageGSX2 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_GSX2];
   } contextSwitch;

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   VA32 hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN32 crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   LA32 monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA32 crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

#ifdef __cplusplus
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 6; }
#endif
} __attribute__ ((packed)) VMCrossPageGSX2;

#define MODULECALL_CROSSPAGE_SIZE_GSX2     0x5dd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX2 0x5b4

/* ---------------------------- WS 3.2.1 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V321	1008

typedef struct VMCrossPageV321 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
   uint32 _pad[3];

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V321];
   } contextSwitch;

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   VA32 hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA32 monitorLinearAddr;      // base location of monitor in linear space
   LA32 crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 4; }
#endif
} __attribute__ ((packed)) VMCrossPageV321;

#define MODULECALL_CROSSPAGE_SIZE_V321     0x5f0
#define MODULECALL_CROSSPAGE_LAOFFSET_V321 0x5c8

/* ---------------------------- WS 3.2 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V32	1008

typedef struct VMCrossPageV32 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
   uint32 _pad[3];

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V32];
   } contextSwitch;

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   VA32 hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA32 monitorLinearAddr;      // base location of monitor in linear space
   LA32 crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 4; }
#endif
} __attribute__ ((packed)) VMCrossPageV32;

#define MODULECALL_CROSSPAGE_SIZE_V32     0x5d0
#define MODULECALL_CROSSPAGE_LAOFFSET_V32 0x5a8

/* ---------------------------- WS 3.0 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V3	192

typedef struct VMCrossPageV3 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V3];
   } contextSwitch;

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   VA32 hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA32 monitorLinearAddr;      // base location of monitor in linear space
   LA32 crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 4; }
#endif
} __attribute__ ((packed)) VMCrossPageV3;

typedef struct VMCrossPageV2 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V3];
   } contextSwitch;

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   VA32 hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA32 monitorLinearAddr;      // base location of monitor in linear space
   LA32 crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 2; }
#endif
} __attribute__ ((packed)) VMCrossPageV2;

typedef struct VMCrossPageGSX1 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   struct {
      uint8 hostXToVmm32[CONTEXT_SWITCH_CODE_SIZE_V3];
   } contextSwitch;

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfoV45 hostContext;
   ContextInfoV45 monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   VA32 hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA32 monitorLinearAddr;      // base location of monitor in linear space
   LA32 crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
   Bool isMCIntr(void) { return moduleCallType == MODULECALL_V3_INTR; }
   static Bool isDRModified(int reg, uint32 savedDR) { return TRUE; }
   static void setDR(int reg, uint32 dr) { }
   static unsigned int crosspageStart(void) { return 2; }
#endif
} __attribute__ ((packed)) VMCrossPageGSX1;

#define MODULECALL_CROSSPAGE_SIZE_V3     0x294
#define MODULECALL_CROSSPAGE_LAOFFSET_V3 0x26c


#define USERCALL_TIMEOUT     100  // milliseconds

#define MX_WAITINTERRUPTED     3
#define MX_WAITTIMEDOUT        2
#define MX_WAITNORMAL          1  // Must equal one; see linux module code.
#define MX_WAITERROR           0  // Use MX_ISWAITERROR() to test for error.

// Any zero or negative value denotes error.
#define MX_ISWAITERROR(e)      ((e) <= MX_WAITERROR)

#define OFF64(_off) offsetof(ContextInfo64,_off)

#endif
